/*
 * Copyright (c) 1995 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Add: at most k packets sent for each ack or new ack.  */


#ifndef lint
static char rcsid[] =
    "@(#) $Header: sack-tcp.cc,v 1.13 95/12/27 14:59:42 floyd Exp $ (LBL)";
#endif

//
// sack-tcp: a selective-ACK capable TCP source
//

#include <stdio.h>
#include <sys/types.h>

#include "packet.h"
#include "tcp.h"
#include "random.h"

#define	PEND_PKTS	100	/* max # of packets to be retransmitted */
#define	PEND_SENTINEL	-1	/* pending packets sentinel */

/* these are just constants in other code (see tcp.cc) */
#define	RECOVER_DUPACK	1
#define	RECOVER_TIMEOUT	2
#define	RECOVER_QUENCH	3

#define	TRUE	1
#define	FALSE	0

/* these are used to mark packets as to why we xmitted them */

#define	REASON_NORMAL	PR_NORMAL
#define	REASON_TIMOUT	PR_REASON1
#define	REASON_DUPACK	PR_REASON2

class SackTcpAgent : public TcpAgent {
public:
	SackTcpAgent();
	int window();
	void recv(Packet *pkt);
	void timeout(int tno);
	void send(int forceflag, int reason, int maxburst = 0);
protected:
	int samax_;		/* max hole just sent */
	int pipe_;		/* estimate of pipe size (fastrecov) */
	int ppend_[PEND_PKTS+1]; /* # of packet to be xmitted w/sentinel */
	int px_;		/* index into ppend_ */
	u_char fastrecov_;	/* boolean: doing fast recovery? */
	u_char timeout_;	/* boolean: sent pkt from timeout? */
};

static class SackTcpMatcher : public Matcher {
public:
	SackTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack") == 0)
			return (new SackTcpAgent());
		return (0);
	}
} matcher_tcpsack;

int SackTcpAgent::window()
{
	int win;

	if (fastrecov_)
                win = 2 * int(wnd_);
        else
                win = int(cwnd_ < wnd_ ? cwnd_ : wnd_);

        if (win > int(2 * wnd_))
                win = int(2 * wnd_);
	return (win);
}

SackTcpAgent::SackTcpAgent() : 
    px_(0), fastrecov_(FALSE), timeout_(FALSE), samax_(-1), pipe_(-1)
{
	ppend_[0] = PEND_SENTINEL;
}

void SackTcpAgent::recv(Packet *pkt)
{
	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}
	register int j = 0, i;
	for (i = 0; pkt->sa_start_[i] >= 0; i++) {
		if (i > NSA) {
			fprintf(stderr,
				"SackTcpAgent::recv: val %d exceeds NSA %d\n",
				i, NSA);
			exit(1);
		}

		for (int k = pkt->sa_start_[i]; k <= pkt->sa_end_[i]; k++) {
			if (k > samax_ && j <= PEND_PKTS - 1) {
				ppend_[j] = k;
				j++;
			}
		}
	}

	if (j <= PEND_PKTS)
		ppend_[j] = PEND_SENTINEL;

	px_ = 0;

	if (pkt->flags_ & PF_ECN)
		quench(1);

	if (!fastrecov_) {
		/* normal... not fast recovery */
		if ((int)pkt->seqno_ > last_ack_) {
			/*
			 * regular ACK not in fast recovery... normal
			 */
			newack(pkt);
			opencwnd();
			timeout_ = FALSE;
		} else if (timeout_ == FALSE)  {
			if (pkt->seqno_ != last_ack_) {
				fprintf(stderr, "pkt seq %d should be = to %d\n",
					pkt->seqno_, last_ack_);
				abort();
			}
			/*
			 * a duplicate ACK
			 */
			if (++dupacks_ == NUMDUPACKS) {
				/*
				 * Assume we dropped just one packet.
				 * Retransmit last ack + 1
				 * and try to resume the sequence.
				 */
				if ((highest_ack_ > recover_) ||
				    (recover_cause_ != RECOVER_TIMEOUT)) {
					recover_cause_ = RECOVER_DUPACK;
					recover_ = maxseq_;
					pipe_ = int(cwnd_) - NUMDUPACKS;
					closecwnd(1);
					reset_rtx_timer(1);
					fastrecov_ = TRUE;
					samax_ = last_ack_ + 1;
					output(samax_, REASON_DUPACK);
				}
			}
		}
		if (dupacks_ == 0)
			send(FALSE, REASON_NORMAL, maxburst_);
	} else {
		/* we are in fast recovery */
		--pipe_;
		if ((int)pkt->seqno_ >= recover_) {
			/* ACK indicates fast recovery is over */
			fastrecov_ = FALSE;
			newack(pkt);
			timeout_ = FALSE;
			/* New window: W/2 - K or W/2? */
		} else if ((int)pkt->seqno_ > highest_ack_) {
			/* Not out of fast recovery yet.
			 * Update highest_ack_, but not last_ack_. */
			highest_ack_ = (int)pkt->seqno_;
			if (pipe_ >= int(cwnd_)) {
				/* 
				 * More that half of the packets in the
				 * window were dropped.
				 */
				/* pipe_ = int(cwnd_) - 1; */
				last_ack_ = (int)pkt->seqno_;
				timeout(TCP_TIMER_RTX);
			}
			/* 
			 * An alternative algorithm would be to assume that
			 * reported holes represent packets that have
			 * left the pipe.  Then we would not have "pipe_"
			 * larger than "cwnd_" when more than half of the
			 * packets in a window were dropped.
			 * Another possibility would be to slow-start here
			 * without waiting for the retransmit timer to
			 * expire.
			 */
			t_backoff_ = 1;
			newtimer(pkt);
		} else if (timeout_ == FALSE) {
			/* got another dup ack */
			if (dupacks_ > 0)
				dupacks_++;
		}
		send(FALSE, REASON_NORMAL, maxburst_);
	}

	Packet::free(pkt);
	if (trace_)
		plot();
}

void SackTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		samax_ = -1;
		ppend_[px_ = 0] = PEND_SENTINEL;
		dupacks_ = 0;
		fastrecov_ = FALSE;
		timeout_ = TRUE;
		if (highest_ack_ > last_ack_)
			last_ack_ = highest_ack_;
		recover_ = maxseq_;
	}
	TcpAgent::timeout(tno);
}

#define	STILLMORE	(seqno_ < curseq_)	/* more app data to send */
void SackTcpAgent::send(int force, int reason, int maxburst)
{
	register int pktno, found, nextpktno, npacket = 0;
	int win = window();

	found = TRUE;
	/*
	 * as long as the pipe is open and there is app data to send...
	 */
	while (((!fastrecov_ && (seqno_ <= last_ack_ + win)) ||
	    (fastrecov_ && (pipe_ < int(cwnd_)))) && STILLMORE && found) {

		if (overhead_ == 0 || force) {
			/*
			 * look for a pending packet to send
			 */
			found = FALSE;
			/* Why do we look in ppend if we are not in
			 * fast recovery? - SF
			 */
			while (((pktno = ppend_[px_]) != PEND_SENTINEL) &&
			    (pktno <= last_ack_ + win) && (found == FALSE)) {

			    if (pktno > samax_) {
				samax_ = pktno;
				if (pktno > last_ack_) {
					nextpktno = pktno;
					found = TRUE;
				}
			    }
			    px_++;
			}
			if (!found && seqno_ > ppend_[px_] &&
			    (seqno_ <= last_ack_ + win)) {
			    /* This is last_ack_, or highest_ack_? 
			     * I think last_ack_, to prevent staying
			     * in Fast Recovery indefinitely.  */
				nextpktno = seqno_++;
				found = TRUE;
			}

			if (found) {
				output(nextpktno, reason);
				npacket++;
				if (seqno_ <= nextpktno)
					seqno_ = nextpktno + 1;
				pipe_++;
			}
		} else if (!pending_[TCP_TIMER_DELSND]) {
			/*
                         * Set a delayed send timeout.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
		}
		if (maxburst && npacket == maxburst)
			break;
	} /* while */
}
