/*
 * Copyright (c) 1991-1995 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: tcp-sink.cc,v 1.24 95/12/27 16:39:18 floyd Exp $ (LBL)";
#endif

#include <math.h>
#include "packet.h"
#include "node.h"
#include "agent.h"

/* max window size */
#define MWS 1024
#define MWM (MWS-1)

class TcpSink : public Agent {
public:
	TcpSink();
	void recv(Packet* pkt);
protected:
	virtual void ack(int seqno, Packet*);
	void update(const Packet*);
	int next_;		/* next packet expected  */
	int maxseen_;		/* max packet number seen */
	int seen_[MWS];		/* array of packets seen  */
};

static class TcpSinkMatcher : public Matcher {
public:
	TcpSinkMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "tcp-sink") == 0)
			return (new TcpSink());
		return (0);
	}
} matcher_tcpsink;

TcpSink::TcpSink() : Agent(PT_ACK)
{
	next_ = 0;
	maxseen_ = 0;
	memset(seen_, 0, sizeof(seen_));
	link_int("ns_sink", "packet-size", &size_, 0);
}

void TcpSink::ack(int seqno, Packet *pkt)
{
	Packet* newpkt = allocpkt(seqno);
	newpkt->flags_ = pkt->flags_;
	newpkt->ts_ = pkt->ts_;
	newpkt->class_ = pkt->class_;

	node_->transmit(newpkt);
}

void TcpSink::update(const Packet* pkt)
{
//      if ((pkt->seqno_ <= maxseen_) && seen_[pkt->seqno_ & MWM])
//        printf ("Packet %d arrived again.\n", pkt->seqno_);

	if (pkt->seqno_ > maxseen_) {
		for (int i = maxseen_ + 1; i < pkt->seqno_; ++i)
			seen_[i & MWM] = 0;
		maxseen_ = pkt->seqno_;
		seen_[maxseen_ & MWM] = 1;
		seen_[(maxseen_ + 1) & MWM] = 0;
	}
	int next = next_;
	if (pkt->seqno_ >= next && pkt->seqno_ <= maxseen_) {
		/*
		 * setting the sequence number.
		 * should be the last in sequence packet seen
		 */
		seen_[pkt->seqno_ & MWM] = 1;
		while (seen_[next & MWM])
			++next;
		next_ = next;
	}
}

void TcpSink::recv(Packet* pkt)
{
	update(pkt);
	ack(next_ - 1, pkt);
	Packet::free(pkt);
}

class DelAckSink : virtual public TcpSink {
public:
	DelAckSink();
	void recv(Packet* pkt);
protected:
	virtual void timeout(int tno);
	Packet* save_;		/* place to stash packet while delaying */
	double interval_;
};

static class DelSinkMatcher : public Matcher {
public:
	DelSinkMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "tcp-sink-da") == 0)
			/*
			return (new TcpSink());
			*/
			return (new DelAckSink());
		return (0);
	}
} matcher_delsink;

DelAckSink::DelAckSink()
{
	link_time("ns_delsink", "interval", &interval_, 0);
}

#define DELAY_TIMER 0

void DelAckSink::recv(Packet* pkt)
{
	update(pkt);
        /*
         * If there's no timer and the packet is in sequence, set a timer.
         * Otherwise, send the ack and update the timer.
         */
        if (!pending_[DELAY_TIMER] && pkt->seqno_ == next_ - 1) {
                /*
                 * There's no timer, so set one and delay this ack.
                 */
		save_ = pkt;
		sched(interval_, DELAY_TIMER);
                return;
        }
        /*
         * If there was a timer, turn it off.
         */
	if (pending_[DELAY_TIMER]) {
		cancel(DELAY_TIMER);
		Packet::free(save_);
		save_ = 0;
	}
	ack(next_ - 1, pkt);
	Packet::free(pkt);
}

void DelAckSink::timeout(int tno)
{
	/*
	 * The timer expired so we ACK the last packet seen.
	 */
	Packet* pkt = save_;
	ack(next_ - 1, pkt);
	Packet::free(pkt);
}

class SackTcpSink : virtual public TcpSink {
	void ack(int, Packet*);
public:
	inline SackTcpSink() { };
};

static class SackTcpSinkMatcher : public Matcher {
public:
	SackTcpSinkMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "sack-tcp-sink") == 0)
			return (new SackTcpSink());
		return (0);
	}
} matcher_sacktcpsink;

void SackTcpSink::ack(int seqno, Packet *pkt)
{

	int delta = 6;
	register i, j;

        Packet* newpkt = allocpkt(seqno);
        newpkt->flags_ = pkt->flags_;
        newpkt->ts_ = pkt->ts_;
        newpkt->class_ = pkt->class_;

	i = next_;
	j = 0;

	for (; (i < maxseen_) && (j < NSA); i++) {
		if (!seen_[i & MWM]) {
			if (seen_[(i-1) & MWM]) {
				newpkt->sa_start_[j] = i;
			}

			if (seen_[(i+1) & MWM]) {
				newpkt->sa_end_[j++] = i;
				newpkt->size_ += delta;
				delta = 4;
			}
		}
	}

	newpkt->sa_start_[j] = -1;
	node_->transmit(newpkt);
}

/* NEW */

class SackDelAckTcpSink : public SackTcpSink, public DelAckSink {
public:
	inline SackDelAckTcpSink() { };
};

static class SackDelAckTcpSinkMatcher : public Matcher {
public:
	SackDelAckTcpSinkMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "sack-tcp-sink-da") == 0)
			return (new SackDelAckTcpSink());
		return (0);
	}
} matcher_sackdelacktcpsink;

/* NEW */
/* "sack1-tcp-sink" is for Matt and Jamshid's implementation of sack. */

#define MAX_OLD_SACK 100

class SackStack {
protected:
	int size_;
	int cnt_;
	struct Sf_Entry {
		int left_;
		int right_;
	} *SFE_;
public:
	SackStack(int);
	~SackStack();
	int& head_right(int n = 0) { return SFE_[n].right_; }
	int& head_left(int n = 0) { return SFE_[n].left_; }
	int cnt() { return cnt_; }
	void reset() {
		register i;
		for (i = 0; i < cnt_; i++)
			SFE_[i].left_ = SFE_[i].right_ = -1;

		cnt_ = 0;
	}

	inline void push(int n = 0) {
		for (register i = cnt_-2; i >= n; i--)
			SFE_[i+1] = SFE_[i];	// not efficient for big size
		cnt_++;
	}

	inline void pop(int n = 0) {
		register i;
		for (i = n; i < cnt_-1; i++)
			SFE_[i] = SFE_[i+1];	// not efficient for big size
		SFE_[i].left_ = SFE_[i].right_ = -1;
		cnt_--;
	}
};

SackStack::SackStack(int sz)
{
	size_ = sz;
	SFE_ = new Sf_Entry[sz];
	for (register i = 0; i < sz; i++)
		SFE_[i].left_ = SFE_[i].right_ = -1;
	cnt_ = 0;
}

SackStack::~SackStack()
{
	delete SFE_;
}

class Sack1TcpSink : virtual public TcpSink {
        void ack(int, Packet*);
public: 
        Sack1TcpSink();
protected:
        int max_sack_blocks_;
	SackStack *sf_;
};

Sack1TcpSink::Sack1TcpSink()
{
        link_int("ns_sacksink", "max-sack-blocks", &max_sack_blocks_, 0);
	sf_ = new SackStack(max_sack_blocks_);
}

Sack1TcpSink::~Sack1TcpSink()
{
	delete sf_;
}

static class Sack1TcpSinkMatcher : public Matcher {
public:
        Sack1TcpSinkMatcher() : Matcher("agent") {}
        TclObject* match(const char* id) {
                if (strcmp(id, "sack1-tcp-sink") == 0)
                        return (new Sack1TcpSink());
                return (0);
        }
} matcher_sack1tcpsink;

void Sack1TcpSink::ack(int seqno, Packet *pkt)
{
        int sack_index, i, sack_right, sack_left;
	int recent_sack_left, recent_sack_right;
          
        Packet* newpkt = allocpkt(seqno);
        newpkt->flags_ = pkt->flags_;
        newpkt->ts_ = pkt->ts_;
        newpkt->class_ = pkt->class_;

        newpkt->sack_length_ = 0;
        sack_index = 0;

#ifdef DEBUGSACK1
printf("ACK:%d, for pkt %d, [maxseen:%d]\n", seqno, pkt->seqno_,
maxseen_);
#endif
        if (pkt->seqno_<0) {
                printf("Error: invalid packet number %d\n", pkt->seqno_);
        }
	else if (seqno >= maxseen_ && (sf_->cnt() != 0))
		sf_->reset();
	else if ((seqno < maxseen_) && (max_sack_blocks_ > 0)) {
                /*  Build FIRST SACK block  */
                sack_right=-1;

		/* look rightward for first hole */
                for (i=pkt->seqno_; i<=maxseen_; i++) {
                              if (!seen_[i & MWM]) {
                                                sack_right=i;
                                                break;
		}
		}

                if (sack_right == -1) {
                                sack_right = maxseen_+1;
                }

		if (pkt->seqno_ <= seqno) {
			sack_left = 0;
		} else {
			/* look leftward from right edge for first hole */
	                for (i = sack_right-1; i > seqno; i--) {
	                                if (!seen_[i & MWM]) {
	                                        sack_left = i+1;
	                                        break;
	                                }
	                }
		}

                newpkt->sa_left_[sack_index] = sack_left;
                newpkt->sa_right_[sack_index] = sack_right;
                newpkt->sack_length_++;
                sack_index++;

		recent_sack_left = sack_left;
		recent_sack_right = sack_right;

		/* first sack block is built, check the others */
		int k = 0;
                while (sack_index < max_sack_blocks_) {

			sack_left = sf_->head_left(k);
			sack_right = sf_->head_right(k);

			/* no more history */
			if (sack_left < 0 || sack_right < 0 ||
				sack_right > maxseen_ + 1)
				break;

			/* newest ack "covers up" this one */
			if (recent_sack_left <= sack_left &&
			    recent_sack_right >= sack_right) {
				sf_->pop(k);
				continue;
			}

			newpkt->sa_left_[sack_index] = sack_left;
			newpkt->sa_right_[sack_index] = sack_right;
			newpkt->sack_length_++;
			sack_index++;
			k++;

                }

		if (pkt->seqno_ > seqno) {
		 	/* put most recent block onto stack */
			sf_->push();
			sf_->head_left() = recent_sack_left;
			sf_->head_right() = recent_sack_right;
		}
                
        }

#ifdef DEBUGSACK1
{
int k;
if (newpkt->sack_length_ > 0) {
  printf("\tpkt %d, [", newpkt->seqno_);
  for (k = 0; k < newpkt->sack_length_; k++)
    printf("(%d,%d),", newpkt->sa_left_[k], newpkt->sa_right_[k]);
  printf("]\n");
}
}
#endif

        node_->transmit(newpkt);
}

/* NEW */

class Sack1DelAckTcpSink : public Sack1TcpSink, public DelAckSink {
public:
	inline Sack1DelAckTcpSink() { };
};

static class Sack1DelAckTcpSinkMatcher : public Matcher {
public:
	Sack1DelAckTcpSinkMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "sack1-tcp-sink-da") == 0)
			return (new Sack1DelAckTcpSink());
		return (0);
	}
} matcher_sack1delacktcpsink;
